package com.cognex.cmbsdk_flutter;

import android.Manifest;
import android.app.Activity;
import android.app.Application;
import android.content.pm.PackageManager;
import android.graphics.Point;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.util.Log;
import android.view.View;
import android.view.ViewGroup;
import android.widget.FrameLayout;
import android.widget.RelativeLayout;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.core.app.ActivityCompat;

import com.cognex.dataman.sdk.CameraMode;
import com.cognex.dataman.sdk.ConnectionState;
import com.cognex.dataman.sdk.DataManDeviceClass;
import com.cognex.dataman.sdk.DataManSystem;
import com.cognex.dataman.sdk.DmccResponse;
import com.cognex.dataman.sdk.PreviewOption;
import com.cognex.dataman.sdk.exceptions.CameraPermissionException;
import com.cognex.mobile.barcode.sdk.ReadResults;
import com.cognex.mobile.barcode.sdk.ReadStringEncoding;
import com.cognex.mobile.barcode.sdk.ReaderDevice;
import com.cognex.mobile.barcode.sdk.Symbology;
import com.manateeworks.BarcodeScanner;
import com.manateeworks.mwoverlay.MWOverlay;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.lang.reflect.Field;

import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import io.flutter.plugin.common.PluginRegistry;

/**
 * CmbsdkFlutterPlugin
 */
public class CmbsdkFlutterPlugin implements FlutterPlugin, MethodCallHandler, ActivityAware {
    private static final String TAG = CmbsdkFlutterPlugin.class.getSimpleName();

    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private MethodChannel channel;

    private EventChannel availabilityChannel;
    private EventChannel connectionStateChannel;
    private EventChannel resultsChannel;
    private EventChannel scanningStateChannel;

    private EventChannel.EventSink availabilityChannelEventSink;
    private EventChannel.EventSink connectionStateChannelEventSink;
    private EventChannel.EventSink resultsChannelEventSink;
    private EventChannel.EventSink scanningStateChannelEventSink;

    private ActivityPluginBinding activityPluginBinding;

    private int param_cameraMode = CameraMode.NO_AIMER;
    private int param_previewOptions = PreviewOption.DEFAULTS;
    private boolean param_isScannerFullScreen = true;
    private int param_triggerType = 2;
    private boolean cmb_stopScanningOnRotate = false;
    private DataManDeviceClass param_deviceType = DataManDeviceClass.MX;

    private RelativeLayout scannerView;

    private int position_xp = 0;
    private int position_yp = 0;
    private int position_wp = 100;
    private int position_hp = 50;

    private int param_positionX = 0;
    private int param_positionY = 0;
    private int param_sizeWidth = 100;
    private int param_sizeHeight = 50;

    private String regKey = null;

    private ReaderDevice readerDevice;

    private boolean isScanning = false;

    // USB Listener, no need for conditional code
    private boolean listeningForUSB = false;

    private final View.OnLayoutChangeListener orientationChangeListener = new View.OnLayoutChangeListener() {
        @Override
        public void onLayoutChange(View v, int left, int top, int right, int bottom, int oldLeft,
                                   int oldTop, int oldRight, int oldBottom) {
            if (param_deviceType != DataManDeviceClass.PhoneCamera) {
                return;
            }

            if (left != oldLeft || top != oldTop || right != oldRight || bottom != oldBottom) {
                // Wait a bit for orientationChangeListener in SDK to be executed
                new Handler(Looper.getMainLooper()).postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        updatePreviewContainerValues();

                        if (isScanning && cmb_stopScanningOnRotate) {
                            toggleScanner(false);
                        }
                    }
                }, 200);
            }
        }
    };

    private static final int CAMERA_REQUEST_PERMISSION_CODE = 234;
    private boolean cameraPermissionRequested = false;
    private final PluginRegistry.RequestPermissionsResultListener permissionsResultListener = new PluginRegistry.RequestPermissionsResultListener() {
        @Override
        public boolean onRequestPermissionsResult(int requestCode, String[] permissions, int[] grantResults) {
            if (requestCode == CAMERA_REQUEST_PERMISSION_CODE) {
                if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    if (Utils.isReaderInit(null, readerDevice)) {
                        readerDevice.connect(null);
                    }
                }
            }

            return false;
        }
    };

    private final Application.ActivityLifecycleCallbacks lifecycleCallbacks = new Application.ActivityLifecycleCallbacks() {
        @Override
        public void onActivityCreated(@NonNull Activity activity, @Nullable Bundle savedInstanceState) {

        }

        @Override
        public void onActivityStarted(@NonNull Activity activity) {
            if (Utils.isReaderInit(null, readerDevice)
                    && readerDevice.getDeviceClass() == DataManDeviceClass.MX
                    && !listeningForUSB) {

                readerDevice.startAvailabilityListening();
                listeningForUSB = true;
            }
        }

        @Override
        public void onActivityResumed(@NonNull Activity activity) {

        }

        @Override
        public void onActivityPaused(@NonNull Activity activity) {

        }

        @Override
        public void onActivityStopped(@NonNull Activity activity) {
            // Up to 2.1.0 MX Connect has bug with availability callback, that's why we need to stop
            // listening when we go in background, and start when we are back
            if (Utils.isReaderInit(null, readerDevice)
                    && readerDevice.getDeviceClass() == DataManDeviceClass.MX
                    && listeningForUSB) {

                readerDevice.stopAvailabilityListening();
                listeningForUSB = false;
            }
        }

        @Override
        public void onActivitySaveInstanceState(@NonNull Activity activity, @NonNull Bundle outState) {

        }

        @Override
        public void onActivityDestroyed(@NonNull Activity activity) {

        }
    };

    @Override
    public void onAttachedToEngine(@NonNull FlutterPluginBinding flutterPluginBinding) {
        channel = new MethodChannel(flutterPluginBinding.getBinaryMessenger(), "cmbsdk_flutter");
        channel.setMethodCallHandler(this);

        availabilityChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "availabilityChannel");
        availabilityChannel.setStreamHandler(new EventChannel.StreamHandler() {
            @Override
            public void onListen(Object arguments, EventChannel.EventSink events) {
                Log.d(TAG, "availabilityChannelEventSink is initialized");
                availabilityChannelEventSink = events;
            }

            @Override
            public void onCancel(Object arguments) {

            }
        });

        connectionStateChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "connectionStateChannel");
        connectionStateChannel.setStreamHandler(new EventChannel.StreamHandler() {
            @Override
            public void onListen(Object arguments, EventChannel.EventSink events) {
                Log.d(TAG, "connectionStateChannelEventSink is initialized");
                connectionStateChannelEventSink = events;
            }

            @Override
            public void onCancel(Object arguments) {

            }
        });

        resultsChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "resultsChannel");
        resultsChannel.setStreamHandler(new EventChannel.StreamHandler() {
            @Override
            public void onListen(Object arguments, EventChannel.EventSink events) {
                Log.d(TAG, "resultsChannelEventSink is initialized");
                resultsChannelEventSink = events;
            }

            @Override
            public void onCancel(Object arguments) {

            }
        });

        scanningStateChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "scanningStateChannel");
        scanningStateChannel.setStreamHandler(new EventChannel.StreamHandler() {
            @Override
            public void onListen(Object arguments, EventChannel.EventSink events) {
                Log.d(TAG, "scanningStateChannelEventSink is initialized");
                scanningStateChannelEventSink = events;
            }

            @Override
            public void onCancel(Object arguments) {

            }
        });
    }

    @Override
    public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {
        channel.setMethodCallHandler(null);

        availabilityChannelEventSink = null;
        availabilityChannel.setStreamHandler(null);

        connectionStateChannelEventSink = null;
        connectionStateChannel.setStreamHandler(null);

        resultsChannelEventSink = null;
        resultsChannel.setStreamHandler(null);

        scanningStateChannelEventSink = null;
        scanningStateChannel.setStreamHandler(null);
    }

    @Override
    public void onAttachedToActivity(@NonNull ActivityPluginBinding binding) {
        Log.i(TAG, TAG + " is attached to activity");
        activityPluginBinding = binding;
        activityPluginBinding.addRequestPermissionsResultListener(permissionsResultListener);
        activityPluginBinding.getActivity().getApplication().registerActivityLifecycleCallbacks(lifecycleCallbacks);
    }

    @Override
    public void onDetachedFromActivity() {
        Log.i(TAG, TAG + " is detached from activity");
        activityPluginBinding.removeRequestPermissionsResultListener(permissionsResultListener);
        activityPluginBinding.getActivity().getApplication().unregisterActivityLifecycleCallbacks(lifecycleCallbacks);
        activityPluginBinding = null;
    }

    @Override
    public void onDetachedFromActivityForConfigChanges() {
    }

    @Override
    public void onReattachedToActivityForConfigChanges(@NonNull ActivityPluginBinding binding) {
    }

    @SuppressWarnings("ConstantConditions")
    @Override
    public void onMethodCall(@NonNull MethodCall call, @NonNull Result result) {
        switch (call.method) {
            case "loadScanner":
                if (activityPluginBinding == null)
                    result.error(ErrorCodes.PLUGIN_DETACHED_FROM_ACTIVITY,
                            TAG + " is detached from activity", null);
                else
                    loadScanner(result, (int) call.arguments());
                return;
            case "connect":
                if (Utils.isReaderInit(result, readerDevice))
                    connect(result);
                return;
            case "disconnect":
                if (Utils.isReaderInit(result, readerDevice))
                    disconnect(result);
                return;
            case "startScanning":
                if (Utils.isReaderInit(result, readerDevice))
                    startScanning(result);
                return;
            case "stopScanning":
                if (Utils.isReaderInit(result, readerDevice))
                    stopScanning(result);
                return;
            case "setSymbologyEnabled":
                if (Utils.isReaderInit(result, readerDevice))
                    setSymbologyEnabled(result, (int) call.argument("symbology"),
                            (boolean) call.argument("enable"));
                return;
            case "isSymbologyEnabled":
                if (Utils.isReaderInit(result, readerDevice))
                    isSymbologyEnabled(result, (int) call.arguments());
                return;
            case "setCameraMode":
                param_cameraMode = call.arguments();
                return;
            case "setPreviewOptions":
                param_previewOptions = call.arguments();
                return;
            case "setPreviewOverlayMode":
                final int overlayMode = call.arguments();
                if (overlayMode == MWOverlay.OverlayMode.OM_CMB.ordinal())
                    MWOverlay.overlayMode = MWOverlay.OverlayMode.OM_CMB;
                else if (overlayMode == MWOverlay.OverlayMode.OM_LEGACY.ordinal())
                    MWOverlay.overlayMode = MWOverlay.OverlayMode.OM_LEGACY;
                return;
            case "setPreviewContainerPositionAndSize":
                setPreviewContainerPositionAndSize((int) call.argument("x"), (int) call.argument("y"),
                        (int) call.argument("width"), (int) call.argument("height"));
                return;
            case "setLightsOn":
                if (Utils.isReaderInit(result, readerDevice))
                    setLightsOn(result, (boolean) call.arguments());
                return;
            case "isLightsOn":
                if (Utils.isReaderInit(result, readerDevice))
                    isLightsOn(result);
                return;
            case "enableImage":
                if (Utils.isReaderInit(result, readerDevice)) {
                    readerDevice.enableImage((boolean) call.arguments());
                    result.success(null);
                }
                return;
            case "enableImageGraphics":
                if (Utils.isReaderInit(result, readerDevice)) {
                    readerDevice.enableImageGraphics((boolean) call.arguments());
                    result.success(null);
                }
                return;
            case "sendCommand":
                if (Utils.isReaderInit(result, readerDevice))
                    sendCommand(result, (String) call.arguments());
                return;
            case "resetConfig":
                if (Utils.isReaderInit(result, readerDevice))
                    resetConfig(result);
                return;
            case "registerSDK":
                regKey = call.arguments();
                return;
            case "getAvailability":
                if (Utils.isReaderInit(null, readerDevice))
                    result.success(readerDevice.getAvailability().ordinal());
                else
                    result.success(ReaderDevice.Availability.UNKNOWN.ordinal());
                return;
            case "getConnectionState":
                if (Utils.isReaderInit(null, readerDevice))
                    result.success(readerDevice.getConnectionState().ordinal());
                else
                    result.success(ConnectionState.Disconnected.ordinal());
                return;
            case "beep":
                if (Utils.isReaderInit(result, readerDevice)) {
                    readerDevice.beep();
                    result.success(null);
                }
                return;
            case "getDeviceBatteryLevel":
                if (Utils.isReaderInit(result, readerDevice))
                    getDeviceBatteryLevel(result);
                return;
            case "getSdkVersion":
                result.success(DataManSystem.getVersion());
                return;
            case "setPreviewContainerFullScreen":
                if (Utils.isReaderInit(result, readerDevice))
                    setPreviewContainerFullScreen(result);
                return;
            case "setStopScannerOnRotate":
                cmb_stopScanningOnRotate = call.arguments();
                return;
            case "enableCameraFlag":
            case "disableCameraFlag":
                if (Utils.isReaderInit(result, readerDevice)) {
                    editCameraFlag(result, (int) call.argument("codeMask"),
                            (int) call.argument("flag"), "enableCameraFlag".equalsIgnoreCase(call.method));
                }
                return;
            case "setParser":
                if (Utils.isReaderInit(result, readerDevice))
                    setParser(result, (int) call.arguments());
                return;
            case "setReadStringEncoding":
                if (Utils.isReaderInit(result, readerDevice))
                    setReadStringEncoding(result, (int) call.arguments());
                return;
            default:
                result.notImplemented();
        }
    }

    private void loadScanner(@NonNull final Result result, int deviceType) {
        param_deviceType = DataManDeviceClass.values()[deviceType];

        removeScannerView();

        if (readerDevice != null) {
            if (listeningForUSB) {
                readerDevice.stopAvailabilityListening();
                listeningForUSB = false;
            }

            readerDevice.setReaderDeviceListener(null);
            readerDevice.disconnect();
            readerDevice = null;

            if (connectionStateChannelEventSink != null)
                connectionStateChannelEventSink.success(ConnectionState.Disconnected.ordinal());
        }

        if (param_deviceType == DataManDeviceClass.PhoneCamera) {
            if (!param_isScannerFullScreen)
                addScannerView();

            readerDevice = ReaderDevice.getPhoneCameraDevice(activityPluginBinding.getActivity(), param_cameraMode,
                    param_previewOptions, param_isScannerFullScreen ? null : scannerView, regKey);
        } else {
            readerDevice = ReaderDevice.getMXDevice(activityPluginBinding.getActivity());

            if (!listeningForUSB) {
                readerDevice.startAvailabilityListening();
                listeningForUSB = true;
            }
        }

        readerDevice.setReaderDeviceListener(new ReaderDevice.ReaderDeviceListener() {
            @Override
            public void onConnectionStateChanged(ReaderDevice reader) {
                if (connectionStateChannelEventSink != null)
                    connectionStateChannelEventSink.success(reader.getConnectionState().ordinal());
            }

            @Override
            public void onReadResultReceived(ReaderDevice reader, ReadResults readResults) {
                if (resultsChannelEventSink != null) {
                    final String xml = readResults.getXml();

                    final JSONArray results = new JSONArray();
                    final JSONArray subResults = new JSONArray();

                    if (readResults.getCount() > 0) {
                        results.put(Utils.readResultToJsonObj(readResults.getResultAt(0)));
                    }

                    if (readResults.getSubResults() != null) {
                        for (int i = 0; i < readResults.getSubResults().size(); i++) {
                            results.put(Utils.readResultToJsonObj(readResults.getSubResults().get(i)));
                            subResults.put(Utils.readResultToJsonObj(readResults.getSubResults().get(i)));
                        }
                    }

                    try {
                        resultsChannelEventSink.success(new JSONObject()
                                .put("xml", xml)
                                .put("results", results)
                                .put("subResults", subResults).toString());
                    } catch (JSONException ignored) {
                    }

                    if (param_triggerType != 5) {
                        // not in continuous scanning, scanner will stop
                        isScanning = false;

                        if (scanningStateChannelEventSink != null)
                            scanningStateChannelEventSink.success(false);
                    }
                }
            }

            @Override
            public void onAvailabilityChanged(ReaderDevice reader) {
                if (availabilityChannelEventSink != null)
                    availabilityChannelEventSink.success(reader.getAvailability().ordinal());
            }
        });

        result.success(null);
    }

    private void connect(@NonNull final Result result) {
        readerDevice.connect(new ReaderDevice.OnConnectionCompletedListener() {
            @Override
            public void onConnectionCompleted(ReaderDevice readerDevice, Throwable throwable) {
                if (throwable != null) {
                    if (throwable instanceof CameraPermissionException) {
                        result.error(ErrorCodes.CONNECTION_CAMERA_PERMISSION, throwable.getMessage(), null);

                        // We will ask for permission only once
                        if (activityPluginBinding != null && !cameraPermissionRequested) {

                            cameraPermissionRequested = true;
                            ActivityCompat.requestPermissions(activityPluginBinding.getActivity(),
                                    new String[]{Manifest.permission.CAMERA}, CAMERA_REQUEST_PERMISSION_CODE);
                        }
                    } else
                        result.error(ErrorCodes.CONNECTION_NOT_SUCCESSFUL, throwable.getMessage(), null);
                } else {
                    result.success(null);
                }
            }
        });
    }

    private void disconnect(@NonNull final Result result) {
        readerDevice.disconnect();

        result.success(null);
    }

    private void startScanning(@NonNull final Result result) {
        if (readerDevice.getConnectionState() == ConnectionState.Connected) {
            toggleScanner(true);

            result.success(null);
        } else {
            result.error(ErrorCodes.SCANNER_NOT_CONNECTED, "Scanner is not connected", null);
        }
    }

    private void stopScanning(@NonNull final Result result) {
        toggleScanner(false);

        result.success(null);
    }

    private void setSymbologyEnabled(@NonNull final Result result, int symbology, boolean enable) {
        final Symbology nativeSymbology = Utils.intToNativeSymbology(symbology);

        if (nativeSymbology == Symbology.UNKNOWN) {
            result.error(ErrorCodes.SYMBOLOGY_NOT_FOUND, "Symbology not found", null);
            return;
        }

        readerDevice.setSymbologyEnabled(nativeSymbology, enable, new ReaderDevice.OnSymbologyListener() {
            @Override
            public void onSymbologyEnabled(ReaderDevice readerDevice, Symbology symbology,
                                           Boolean aBoolean, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.SYMBOLOGY_STATE_NOT_CHANGED, throwable.getMessage(), null);
                else
                    result.success(null);
            }
        });
    }

    private void isSymbologyEnabled(@NonNull final Result result, int symbology) {
        final Symbology nativeSymbology = Utils.intToNativeSymbology(symbology);

        if (nativeSymbology == Symbology.UNKNOWN) {
            result.error(ErrorCodes.SYMBOLOGY_NOT_FOUND, "Symbology not found", null);
            return;
        }

        readerDevice.isSymbologyEnabled(nativeSymbology, new ReaderDevice.OnSymbologyListener() {
            @Override
            public void onSymbologyEnabled(ReaderDevice readerDevice, Symbology symbology,
                                           Boolean aBoolean, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.SYMBOLOGY_STATE_NOT_AVAILABLE, throwable.getMessage(), null);
                else
                    result.success(aBoolean);
            }
        });
    }

    private void setPreviewContainerPositionAndSize(int x, int y, int width, int height) {
        position_xp = x;
        position_yp = y;
        position_wp = width;
        position_hp = height;

        param_isScannerFullScreen = false;

        updatePreviewContainerValues();

        try {
            if (Utils.isReaderInit(null, readerDevice) && scannerView != null)
                readerDevice.setCameraPreviewContainer(scannerView);
        } catch (UnsupportedOperationException ignored) {
        }
    }

    private void setLightsOn(@NonNull final Result result, boolean on) {
        readerDevice.setLightsOn(on, new ReaderDevice.OnLightsListener() {
            @Override
            public void onLightsOnCompleted(ReaderDevice readerDevice, Boolean aBoolean, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.LIGHT_STATE_NOT_CHANGED, throwable.getMessage(), null);
                else
                    result.success(null);
            }
        });
    }

    private void isLightsOn(@NonNull final Result result) {
        readerDevice.isLightsOn(new ReaderDevice.OnLightsListener() {
            @Override
            public void onLightsOnCompleted(ReaderDevice readerDevice, Boolean aBoolean, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.LIGHT_STATE_NOT_AVAILABLE, throwable.getMessage(), null);
                else
                    result.success(aBoolean);
            }
        });
    }

    private void sendCommand(@NonNull final Result result, String commandString) {
        readerDevice.getDataManSystem().sendCommand(commandString, new DataManSystem.OnResponseReceivedListener() {
            @Override
            public void onResponseReceived(DataManSystem dataManSystem, DmccResponse dmccResponse) {
                if (dmccResponse.getError() != null)
                    result.error(ErrorCodes.COMMAND_NOT_EXECUTED, dmccResponse.getError().getMessage(), null);
                else
                    result.success(dmccResponse.getPayLoad());
            }
        });
    }

    private void resetConfig(@NonNull final Result result) {
        readerDevice.resetConfig(new ReaderDevice.OnResetConfigListener() {
            @Override
            public void onResetConfigCompleted(ReaderDevice readerDevice, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.CONFIG_NOT_RESETED, throwable.getMessage(), null);
                else
                    result.success(null);
            }
        });
    }

    private void getDeviceBatteryLevel(@NonNull final Result result) {
        readerDevice.getDeviceBatteryLevel(new ReaderDevice.OnDeviceBatteryLevelListener() {
            @Override
            public void onDeviceBatteryLevelReceived(ReaderDevice readerDevice, int batteryLevel, Throwable throwable) {
                if (throwable != null)
                    result.error(ErrorCodes.BATTERY_LEVEL_NOT_AVAILABLE, throwable.getMessage(), null);
                else
                    result.success(batteryLevel);
            }
        });
    }

    private void setPreviewContainerFullScreen(@NonNull final Result result) {
        try {
            readerDevice.setCameraPreviewContainer(null);
            param_isScannerFullScreen = true;
            result.success(null);
        } catch (UnsupportedOperationException ex) {
            result.error(ErrorCodes.UNSUPPORTED_OPERATION, ex.getMessage(), null);
        }
    }

    private void editCameraFlag(@NonNull final Result result, int mask, int flag, boolean enable) {
        int executionResult;
        if (enable)
            executionResult = BarcodeScanner.MWBenableFlag(mask, flag);
        else
            executionResult = BarcodeScanner.MWBdisableFlag(mask, flag);

        switch (executionResult) {
            case BarcodeScanner.MWB_RT_OK:
                result.success(null);
                return;
            case BarcodeScanner.MWB_RT_NOT_SUPPORTED:
                result.error(ErrorCodes.CAMERA_FLAG_NOT_SUPPORTED,
                        "Flag value not supported for selected decoder", null);
                return;
            case BarcodeScanner.MWB_RT_BAD_PARAM:
                result.error(ErrorCodes.CAMERA_FLAG_BAD_PARAM,
                        "Flag value out of range", null);
                return;
            default:
                result.error(ErrorCodes.CAMERA_FLAG_NOT_CHANGED, "Unknown", null);
        }
    }

    private void setParser(@NonNull final Result result, int parser) {
        final ReaderDevice.ResultParser resultParser = Utils.intToParser(parser);

        if (resultParser == null) {
            result.error(ErrorCodes.PARSER_NOT_FOUND, "Parser not found", null);
            return;
        }

        readerDevice.setParser(resultParser);
        result.success(null);
    }

    private void setReadStringEncoding(@NonNull final Result result, int encoding) {
        final ReadStringEncoding readStringEncoding = Utils.intToReadStringEncoding(encoding);

        if (readStringEncoding == null) {
            result.error(ErrorCodes.READ_STRING_ENCODING_NOT_FOUND, "Encoding not found", null);
            return;
        }

        readerDevice.setReadStringEncoding(readStringEncoding);
        result.success(null);
    }

    //region Helper functions

    private void toggleScanner(boolean scan) {
        if (scan) {
            startScan();
        } else {
            stopScan();
        }

        isScanning = scan;
    }

    private void startScan() {
        readerDevice.getDataManSystem().sendCommand("GET TRIGGER.TYPE", new DataManSystem.OnResponseReceivedListener() {
            @Override
            public void onResponseReceived(DataManSystem dataManSystem, DmccResponse dmccResponse) {
                if (dmccResponse.getError() == null)
                    param_triggerType = Integer.parseInt(dmccResponse.getPayLoad());
            }
        });

        readerDevice.startScanning();

        if (scanningStateChannelEventSink != null)
            scanningStateChannelEventSink.success(true);
    }

    private void stopScan() {
        readerDevice.stopScanning();

        if (scanningStateChannelEventSink != null)
            scanningStateChannelEventSink.success(false);
    }

    private void removeScannerView() {
        if (scannerView == null) {
            return;
        }

        final ViewGroup parent = (ViewGroup) scannerView.getParent();

        if (parent != null) {
            parent.removeView(scannerView);
            parent.removeOnLayoutChangeListener(orientationChangeListener);
        }

        scannerView = null;
    }

    private void addScannerView() {
        if (scannerView == null) {
            scannerView = new RelativeLayout(activityPluginBinding.getActivity());
            scannerView.setFocusableInTouchMode(false);
            scannerView.setFocusable(false);
        }

        if (scannerView.getParent() != null) {
            updatePreviewContainerValues();
        } else {
            final ViewGroup parent = activityPluginBinding.getActivity().findViewById(android.R.id.content);

            if (parent != null) {
                parent.addView(scannerView);
                updatePreviewContainerValues();

                parent.addOnLayoutChangeListener(orientationChangeListener);
            }
        }
    }

    private void updatePreviewContainerValues() {
        if (scannerView == null)
            return;

        final Point size = new android.graphics.Point(((ViewGroup) scannerView.getParent()).getWidth(),
                ((ViewGroup) scannerView.getParent()).getHeight());

        param_positionX = (int) Math.floor(position_xp / 100.0f * size.x);
        param_positionY = (int) Math.floor(position_yp / 100.0f * size.y);

        param_sizeWidth = (int) Math.floor(position_wp / 100.0f * size.x);
        param_sizeHeight = (int) Math.floor(position_hp / 100.0f * size.y);

        updateScannerViewPosition();
    }

    @SuppressWarnings("rawtypes")
    private void updateScannerViewPosition() {
        if (scannerView == null)
            return;

        try {
            if (scannerView.getLayoutParams() != null) {

                scannerView.getLayoutParams().width = param_sizeWidth;
                scannerView.getLayoutParams().height = param_sizeHeight;

                if (scannerView.getLayoutParams().getClass() == FrameLayout.LayoutParams.class) {
                    final FrameLayout.LayoutParams params = (FrameLayout.LayoutParams) scannerView.getLayoutParams();
                    params.setMargins(param_positionX, param_positionY, 0, 0);
                    params.gravity = android.view.Gravity.TOP;

                    scannerView.setLayoutParams(params);
                } else {
                    final Class c = scannerView.getLayoutParams().getClass();
                    Field field = null;

                    try {
                        field = c.getDeclaredField("x");
                    } catch (Exception ignored) {
                    }

                    try {
                        field = c.getDeclaredField("leftMargin");
                    } catch (Exception ignored) {
                    }

                    try {
                        field = c.getDeclaredField("marginLeft");
                    } catch (Exception ignored) {
                    }

                    try {
                        if (field != null) {
                            field.setAccessible(true);
                            field.set(scannerView.getLayoutParams(), param_positionX);
                        }
                    } catch (Exception ignored) {
                    }

                    try {
                        field = c.getDeclaredField("y");
                    } catch (Exception ignored) {
                    }
                    try {
                        field = c.getDeclaredField("topMargin");
                    } catch (Exception ignored) {
                    }
                    try {
                        field = c.getDeclaredField("marginTop");
                    } catch (Exception ignored) {
                    }

                    try {
                        if (field != null) {
                            field.setAccessible(true);
                            field.set(scannerView.getLayoutParams(), param_positionY);
                        }
                    } catch (Exception ignored) {
                    }
                }
            } else {
                final ViewGroup.MarginLayoutParams scannerViewParams =
                        new ViewGroup.MarginLayoutParams(param_sizeWidth, param_sizeHeight);

                scannerViewParams.leftMargin = Math.round(param_positionX);
                scannerViewParams.topMargin = Math.round(param_positionY);
                scannerView.setLayoutParams(scannerViewParams);
            }
        } catch (Exception ignored) {
        }
    }

    //endregion
}
