package com.cognex.cmbsdk_flutter;

import android.graphics.Bitmap;
import android.util.Base64;

import com.cognex.mobile.barcode.sdk.ReadResult;
import com.cognex.mobile.barcode.sdk.ReadStringEncoding;
import com.cognex.mobile.barcode.sdk.ReaderDevice;
import com.cognex.mobile.barcode.sdk.Symbology;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.ByteArrayOutputStream;

import io.flutter.plugin.common.MethodChannel;

public class Utils {

    static JSONObject readResultToJsonObj(ReadResult result) {

        int symbology = -1;
        String symbologyString = "NO READ";
        String readString = "";
        final boolean goodRead;
        String xml = null;
        String imageGraphics = null;
        String image = null;
        String parsedText = null;
        String parsedJSON = null;
        final boolean isGS1;

        if (result.isGoodRead()) {
            if (result.getSymbology() != null) {
                symbology = result.getSymbology().ordinal();
                symbologyString = result.getSymbology().getName();
            }
            readString = result.getReadString();
        }

        goodRead = result.isGoodRead();

        if (result.getXml() != null) {
            xml = result.getXml();
        }

        if (result.getImageGraphics() != null) {
            imageGraphics = result.getImageGraphics();
        }

        if (result.getImage() != null) {
            try (ByteArrayOutputStream byteArrayOutputStream = new java.io.ByteArrayOutputStream()) {
                result.getImage().compress(Bitmap.CompressFormat.PNG, 100, byteArrayOutputStream);
                final byte[] byteArray = byteArrayOutputStream.toByteArray();

                image = Base64.encodeToString(byteArray, android.util.Base64.DEFAULT);
            } catch (Exception ignored) {
            }
        }

        if (result.getParsedText() != null)
            parsedText = result.getParsedText();

        if (result.getParsedJSON() != null)
            parsedJSON = result.getParsedJSON();

        isGS1 = result.getIsGS1();

        try {
            return new JSONObject()
                    .put("symbology", symbology)
                    .put("symbologyString", symbologyString)
                    .put("readString", readString)
                    .put("goodRead", goodRead)
                    .put("isGS1", isGS1)
                    .put("parsedText", parsedText)
                    .put("parsedJSON", parsedJSON)
                    .put("imageGraphics", imageGraphics)
                    .put("image", image)
                    .put("xml", xml);
        } catch (JSONException ignored) {
            return new JSONObject();
        }
    }

    static Symbology intToNativeSymbology(int symbology) {
        try {
            return Symbology.values()[symbology];
        } catch (ArrayIndexOutOfBoundsException ignored) {
            return Symbology.UNKNOWN;
        }
    }

    static ReaderDevice.ResultParser intToParser(int parser) {
        try {
            return ReaderDevice.ResultParser.values()[parser];
        } catch (ArrayIndexOutOfBoundsException ignored) {
            return null;
        }
    }

    static ReadStringEncoding intToReadStringEncoding(int encoding) {
        try {
            return ReadStringEncoding.values()[encoding];
        } catch (ArrayIndexOutOfBoundsException ignored) {
            return null;
        }
    }

    static boolean isReaderInit(MethodChannel.Result result, ReaderDevice readerDevice) {
        final boolean isInit = readerDevice != null;

        if (!isInit && result != null)
            result.error(ErrorCodes.SCANNER_NOT_INITIALIZED, "Scanner is not initialized", null);

        return isInit;
    }
}
